#include <stdio.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/resource.h>

void doit(time_t start, int id, char *a);
char *salloc(int n);

#define GRANULE 100 /* microseconds per sample */
#define SAMPLES 20000

main(int argc, char *argv[])
{
  time_t now;
  int i, j, np = 1, pid;
  char *aa[100];

  if(argc == 2){
    np = atoi(argv[1]);
  } else if(argc != 1){
    fprintf(stderr, "Usage: csmeter [number-of-processes]\n");
    exit(1);
  }

  if(np < 1){
    fprintf(stderr, "csmeter: less than one process!\n");
    exit(1);
  }

  /* Reduce scheduling priority. */
  if(setpriority(PRIO_PROCESS, getpid(), 18) < 0){
    perror("csmeter: setpriority");
    exit(1);
  }

  time(&now);

  for(i = 0; i < np-1; i++){
    aa[i] = salloc(SAMPLES);
    pid = fork();
    if(pid < 0){
      perror("csmeter: fork");
      exit(1);
    }
    if(pid == 0){
      doit(now + 2, i + 1, aa[i]);
      exit(0);
    }
  }

  aa[np-1] = salloc(SAMPLES);
  doit(now + 2, np, aa[np-1]);

  for(i = 0; i < SAMPLES; i++){
    int any = 0;
    for(j = 0; j < np; j++){
      if(aa[j][i]){
        printf("%.2f %d\n", (i*GRANULE)/1000.0, j + 1);
        any = 1;
      }
    }
    if(any == 0)
      printf("%.2f 0\n", (i*GRANULE)/1000.0);
  }

  exit(0);
}

void
doit(time_t start, int id, char *a)
{
  struct timeval tv, ntv;
  long long d;
  int ng, dfd;

  while(1){
    gettimeofday(&tv, (struct timezone *) 0);
    if(tv.tv_sec >= start)
      break;
    usleep(10000);
  }

  while(1){
    gettimeofday(&ntv, (struct timezone *) 0);
    d = (ntv.tv_sec - start) * (long long) 1000000
      + ntv.tv_usec;
    ng = d / GRANULE;
    if(ng >= SAMPLES)
      break;
    a[ng] = id;
  }
}

char *
salloc(int n)
{
  char tfn[100];
  int tfd;
  void *p;
  char *a;
  int i;
  char dummy[512];

  sprintf(tfn, "/tmp/csmeter-XXXXXX");
  tfd = mkstemp(tfn);
  if(tfd < 0){
    perror("csmeter: mkstemp");
    exit(1);
  }
  unlink(tfn);
  for(i = 0; i < n; i += sizeof(dummy)){
    if(write(tfd, dummy, sizeof(dummy)) != sizeof(dummy)){
      perror("csmeter: write dummy");
      exit(1);
    }
  }
  p = mmap((void*) 0, n, PROT_READ|PROT_WRITE, MAP_SHARED,
           tfd, 0);
  if(p == MAP_FAILED){
    fprintf(stderr, "csmeter: mmap() failed\n");
    exit(1);
  }
  close(tfd);
  a = (char *) p;
  for(i = 0; i < n; i++)
    a[i] = 0;

  return(a);
}
