package ex3;

import java.util.*;

public class Graph {

  HashMap NodeMap;

  
  public Graph() {
    NodeMap = new HashMap();
  }


  // adds a node to the current graph.
  // @requires: Node o does not already exist
  // @return: true if the node is successfully added, false if it's
  // already exists
  // @modifies this
  public boolean addNode(Object o) {
    if (NodeMap.containsKey(o)) {
      return false;
    }
    else {

      // inserts the new node with no edges.
      NodeMap.put(o, new EdgeSet());
      return true;
    }
  }


  // adds an edge
  // @return false if the origin of the edge doesn't exist
  // or if the edge is already there.
  public boolean addEdge(Edge e) {
    EdgeSet oldset, newset;
    
    if (NodeMap.containsKey(e.origin())) {
      oldset = (EdgeSet) NodeMap.get(e.origin());
      if (oldset.contains(e)) {
	return false;
      }
      else {
      oldset.add(e);
      NodeMap.put(e.origin(), oldset);
      return true;
      }
    }
    else return false;
  }

  // gets all the nodes from the graph. modifications to the set view
  // will be reflected in the graph.
  // @return set view of nodes in graph.
  public Set getNodes() {
    return NodeMap.keySet();
  }

  public boolean containsNode(Object n) {
    return NodeMap.containsKey(n);
  }
  
  // @return set view of edges in Graph, null if node doesn't exist.   
  public EdgeSet getEdges(Object n) {
    if (this.containsNode(n)) {
      return (EdgeSet) NodeMap.get(n);
    }
    else return null;
  }
  
  public int nodeSize() {
    return NodeMap.size();
  }


  // iterate over all the nodes to count the edges per node.
  public int edgeSize() {
    Object curnode;
    Set keys = this.getNodes();
    int count = 0;
    
    for (Iterator it = keys.iterator(); it.hasNext(); ) {
      curnode = it.next();
      count += this.getEdges(curnode).size();
    }
    return count;
  }

  // removes node n and all of the edges that come from it.
  public boolean removeNode(Object n) {
    Set keys;
    EdgeSet oldset, newset = new EdgeSet();
    Object curnode;
    Edge thisedge;
    
    if (NodeMap.containsKey(n)) {
      NodeMap.remove(n);
      keys = this.getNodes();

      for (Iterator it = keys.iterator(); it.hasNext(); ) {
	curnode = it.next();
	oldset = (EdgeSet) NodeMap.get(curnode);
	newset.removeAll(newset);
	for (Iterator edgeit = oldset.iterator(); edgeit.hasNext(); ) {
	  thisedge = (Edge) edgeit.next();
	  if (thisedge.dest() != n) {
	    newset.add(thisedge);
	  }
	}
	NodeMap.put(curnode, newset);
      }
      return true;
    }
    else return false;
  }

  // returns false if: edge does not exist.
  public boolean removeEdge(Edge e) {
    EdgeSet oldset, newset;
    
    if (NodeMap.containsKey(e.origin())) {
      oldset = (EdgeSet) NodeMap.get(e.origin());
      if (oldset.contains(e)) {
	oldset.remove(e);
	NodeMap.put(e.origin(), oldset);
	return true;
      }
    }
    return false;
  }
  
}



