package ex6.ui;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.util.*;
import javax.swing.table.AbstractTableModel;
import ex6.backend.PortfolioActionHandler;
import ex6.backend.PortfolioView;
import ex6.backend.MyPortfolioActionHandler;
import ex6.backend.MyPortfolioView;
import ex6.NoSuchTickerException;
import ex6.WebsiteDataException;

/* FolioTracker
 * a window with menus, including an exit button that works.
 * now you can add portfolios.
 */

public class FolioTracker extends JFrame 
                      implements ActionListener, ItemListener {
  JTextArea output;
  JScrollPane scrollPane;
  String newline = "\n";
  JPanel cards;
  TextEntryDialog FolioNameDialog;
  ListDialog DeleteDialog;
  StockEntryDialog StockDialog;
  Vector FolioViews;
  PortfolioActionHandler FolioActionHandler;
  JMenu stockMenu;
  int currentview; // -1 when overview, non negative otherwise.

  final static int OVERVIEW = -1;
  
  OVTableModel ovModel;
  //PVTableModel pvModels;
  Vector pvTables; // Vector[JTable]
  
 final static String[][] EMPTYOV = {{"",""}};
 final static String[][] EMPTYPV = {{"","","",""}};

// this is where we create the table model for the overview
 class OVTableModel extends AbstractTableModel {

        final String[] columnNames = {"Portfolio", 
                                      "Total Value"};
        final Object[][] data = EMPTYOV;

        public int getColumnCount() {
            return columnNames.length;
        }
        
        public int getRowCount() {
	  if (FolioViews == null) return 0;
	  return FolioViews.size();
        }

        public String getColumnName(int col) {
            return columnNames[col];
        }

        public Object getValueAt(int row, int col) {
	  if (getColumnName(col).equals("Portfolio"))
	    return ((PortfolioView) FolioViews.elementAt(row)).getName();
	  else if (getColumnName(col).equals("Total Value"))
	    return ((PVTableModel) pvTables.elementAt(row)).getTotalValue();
	  return "";
          
        }

   public synchronized void updateTable() {
     fireTableDataChanged();
  }
   
        /*
         * JTable uses this method to determine the default renderer/
         * editor for each cell.  If we didn't implement this method,
         * then the last column would contain text ("true"/"false"),
         * rather than a check box.
         */
        public Class getColumnClass(int c) {
            return getValueAt(0, c).getClass();
        }

    }

// end of OVTableModel

// PVTableModel
// this is where we create the PortfolioView table model
 class PVTableModel extends AbstractTableModel {

   final String[] columnNames = {"Ticker", 
				 "Shares Held",
				 "Price per share",
				 "Last Tick",
				 "Value"};
   final Object[][] data = EMPTYPV;
   final PortfolioView folio;

   public Float getTotalValue() {
     float totalvalue = 0;
     for (int r = 0; r < getRowCount(); r++) {
       
       totalvalue += ((Float) getValueAt(r, 4)).floatValue();
     }
     return new Float(totalvalue);
   }
   
   public PVTableModel(PortfolioView pf) {
     folio = pf;
   }
   
   public int getColumnCount() {
     return columnNames.length;
   }
   
   public int getRowCount() {
     return folio.numStocks();
   }
   
   public String getColumnName(int col) {
     return columnNames[col];
   }
   
   public Object getValueAt(int row, int col) {
     if (getColumnName(col).equals("Ticker"))
       return folio.getTicker(row);
     if (getColumnName(col).equals("Shares Held"))
       return new Integer(folio.getShares(row));
     if (getColumnName(col).equals("Price per share"))
       return new Float(folio.getPrice(row));
     if (getColumnName(col).equals("Value")) {
       int tempshares = ((Integer) getValueAt(row, 1)).intValue();
       float tempprice = ((Float) getValueAt(row, 3)).floatValue();
       return new Float(tempshares * tempprice);
     }
     if (getColumnName(col).equals("Last Tick"))
       return new Float(getTickerPrice(folio.getTicker(row)));
     
     return "50";
   }

   // a wrapper for the FolioActionHandler method.
   public float getTickerPrice(String ticker) {
     float price = 0;

     try {
       price =  FolioActionHandler.getTickerPrice(ticker);
     } catch (NoSuchTickerException e) {
       errorWindow(e.getMessage());
     } catch (WebsiteDataException e) {
       errorWindow(e.getMessage());
     }

     return price;
   }


   void errorWindow(String error) {
     System.out.println("error");
   }
   
   public synchronized void updateTable() {
     fireTableDataChanged();
  }
   /*
    * JTable uses this method to determine the default renderer/
    * editor for each cell.  
    */
   public Class getColumnClass(int c) {
     return getValueAt(0, c).getClass();
   }
   
 }
  // end of PVTableModel
  
  // menus created here.
  void makemenus() {
    JMenuBar menuBar;
    JMenu menu, submenu;
    JMenuItem menuItem;
    
    //Create the menu bar.
    menuBar = new JMenuBar();
    setJMenuBar(menuBar);
    
    //Build the first menu.
    menu = new JMenu("Portfolio");
    menu.setMnemonic(KeyEvent.VK_P);
    menu.getAccessibleContext().setAccessibleDescription(
							 "Open or Close Portfolios");
    menuBar.add(menu);
    
    //a group of JMenuItems
    menuItem = new JMenuItem("New Portfolio",
			     KeyEvent.VK_N);
    menuItem.setAccelerator(KeyStroke.getKeyStroke(
						   KeyEvent.VK_N, ActionEvent.ALT_MASK));
    menuItem.getAccessibleContext().setAccessibleDescription(
							     "Create a new empty Portfolio");
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    menuItem = new JMenuItem("Open Portfolio", 
			     KeyEvent.VK_O);
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    menuItem = new JMenuItem("Close Portfolio");
    menuItem.setMnemonic(KeyEvent.VK_C);
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    menuItem = new JMenuItem("Delete Portfolio");
    menuItem.setMnemonic(KeyEvent.VK_D);
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    menuItem = new JMenuItem("Exit");
    menuItem.setMnemonic(KeyEvent.VK_X);
    menuItem.addActionListener(this);
    menu.add(menuItem);
    
    //Build second menu in the menu bar.
    
    menu = new JMenu("View");
    menu.setMnemonic(KeyEvent.VK_V);
    menu.getAccessibleContext().setAccessibleDescription("Change the view of Portfolios");
    menuBar.add(menu);
    
    menuItem = new JMenuItem("All Portfolios", 
			     KeyEvent.VK_A);
    menuItem.addActionListener(this);
    menu.add(menuItem);

    // make the stock menu, disable it, and add it.
    stockMenu = new JMenu("Stock");
    menuItem = new JMenuItem("Add Stock", KeyEvent.VK_A);
    menuItem.addActionListener(this);
    stockMenu.add(menuItem);
    
    menuItem = new JMenuItem("Delete Stock", KeyEvent.VK_D);
    menuItem.addActionListener(this);
        stockMenu.add(menuItem);

	menuItem = new JMenuItem("Edit Stock", KeyEvent.VK_E);
	menuItem.addActionListener(this);
        stockMenu.add(menuItem);

	stockMenu.setEnabled(false);
	menuBar.add(stockMenu);
	
 }
// end void makemenus()

// FolioTracker constructor
// sets up the basic window.
    public FolioTracker() {
 
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });

	makemenus();

        //Add regular components to the window, using the default BorderLayout.
        Container contentPane = getContentPane();

	cards = new JPanel();
	cards.setLayout(new CardLayout());

	// now for the table
	//OVTableModel
	ovModel = new OVTableModel();
	JTable ovtable = new JTable(ovModel);
        //ovtable.setPreferredScrollableViewportSize(new Dimension(500, 70));
	
	// ovTable panel
	JPanel ovpanel = new JPanel();
	ovpanel.add(ovtable);
	
        //Create the scroll pane and add the table to it. 
	JScrollPane ovscrollPane = new JScrollPane(ovtable);
	
	cards.add(ovscrollPane, "All Portfolios");
	
        //Add the scroll pane to this window.
        contentPane.add(cards, BorderLayout.CENTER);
	int currentview = OVERVIEW;
	
	// make dialogs!
	
	FolioNameDialog = new TextEntryDialog(this, "Name of new portfolio:",
					      "New Portfolio", this);
	FolioNameDialog.pack();
	
	StockDialog = new StockEntryDialog("Enter New Stock", this);
	StockDialog.pack();
	
	// create folio views
	FolioViews = new Vector();
	
	FolioActionHandler = new MyPortfolioActionHandler();
	pvTables = new Vector();
	
 }
  
  public void actionPerformed(ActionEvent e) {
      int i = 0; 
      
      JMenuItem source = (JMenuItem)(e.getSource());
      String s = "Action event detected."
	+ newline
	+ "    Event source: " + source.getText()
	+ " (an instance of " + getClassName(source) + ")";
      
      // beginning of functionality
      
      String sourcetext = source.getText();
      
      // this should not be if/else, it should be switch!
      
      if(source.getText() == "Exit") System.exit(0);
      else if(sourcetext.equals("New Portfolio"))
	createPortfolio();
      else if(sourcetext.equals("Close Portfolio"))
	closePortfolio();
      else if(sourcetext.equals("All Portfolios")) {
	displayPanel("All Portfolios", 0);
	currentview = OVERVIEW;
      }
      else if(Character.isDigit(sourcetext.charAt(0))) {
	do
	  i++;
	while (Character.isDigit(sourcetext.charAt(i)));
	Integer num = Integer.valueOf(sourcetext.substring(0, i));
	int num2 = num.intValue();
	currentview = num2 - 1;
	displayPanel((String)sourcetext.substring(i+1), num2 - 1);
      }
      else if(sourcetext.equals("Add Stock"))
	addStock();
      else if(sourcetext.equals("Delete Stock"))
	deleteStock();
      else if (sourcetext.equals("Edit Stock"))
	editStock();

  } 

  void addStock() {
    StockDialog.setLocationRelativeTo(this);
    StockDialog.setVisible(true);
    String[] stockinfo = StockDialog.getValidatedText();
    addStockToPortfolio(getCurrentPortfolio(), stockinfo[0],
			(Integer.valueOf(stockinfo[1])).intValue(),
			(Float.valueOf(stockinfo[2])).floatValue());
    refreshView();

    
  }

  void deleteStock() {
  }

  void editStock() {
    StockDialog.initialize("Edit Stock", blah, blah, blah);
    StockDialog.setLocationRelativeTo(this);
    StockDialog.setVisible(true);
    String[] stockinfo = StockDialog.getValidatedText();
    addStockToPortfolio(getCurrentPortfolio(), stockinfo[0],
			(Integer.valueOf(stockinfo[1])).intValue(),
			(Float.valueOf(stockinfo[2])).floatValue());
    refreshView();
  }
  
  void addStockToPortfolio(String folioname, String ticker, int shares, float price) {
    FolioActionHandler.addStock(ticker, shares, price, folioname);
  }


  void refreshView() {
  }

  String getCurrentPortfolio() {
    return ((PortfolioView)FolioViews.elementAt(currentview)).getName();
  }
  
  void closePortfolio() {
    
    removeFolio(getCurrentPortfolio());
    
  }
 

  /*  void deletePortfolio() {
    String[] portfoliolist = {"one", "two", "three"};
    DeleteDialog.initialize(this, portfoliolist, "Delete Portfolio", 
			    "Select the portfolio to be deleted:", "Delete");
    String delFolioName = ListDialog.showDialog(null, null);
    
    if (delFolioName != null) {
      // delete from file here.
      
      
      removeFolio(delFolioName);
    }
    }*/
  
 // Removes the portfolio with name folioName from the window of the program.
  void removeFolio(String folioName) {
    
    
  }
  
  void createPortfolio() {
    FolioNameDialog.setLocationRelativeTo(this);
    FolioNameDialog.setVisible(true);
    
    String folioName = FolioNameDialog.getValidatedText();
    if (folioName != null) {    

      PortfolioView newview = FolioActionHandler.createPortfolio(folioName); 
      FolioViews.add(newview);
      
      addFolioToMenu(folioName, FolioViews.size());
      PVTableModel pvModel = new PVTableModel(newview);
      createFolioPanel(pvModel, folioName);
    }
  }
  
  void addFolioToMenu(String folioName, int index) {
    JMenuBar menubar = getJMenuBar();
    JMenu menu = menubar.getMenu(1);
    JMenuItem menuItem = new JMenuItem(index + " " + folioName);
    
    menuItem.addActionListener(this);
    menu.add(menuItem);
  }
  
  
    boolean displayPanel(String NameofPanel, int index) {
      CardLayout cl = (CardLayout)(cards.getLayout());
      if (NameofPanel.equals("All Portfolios")) {	
	stockMenu.setEnabled(false);
	ovModel.updateTable();
      }
      else {
	stockMenu.setEnabled(true);
	((PVTableModel)(((JTable)pvTables.elementAt(index)).getModel())).updateTable();
      }
      cl.show(cards, NameofPanel);
      this.setTitle("FolioTracker - " + NameofPanel);

      return true;
    }

    void createFolioPanel(PVTableModel pvModel, String folioName) {

      JTable pvtable = new JTable(pvModel);
      pvTables.add(pvtable);
      pvtable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

      //      ListSelectionModel rowSM = table.getSelectionModel();
      // int selectedRow = lsm.getMinSelectionIndex();      
      
      //Create the scroll pane and add the table to it. 
      JScrollPane pvscrollPane = new JScrollPane(pvtable);
      //addFolioToMenu(folioName, pvModelVect.size());
      cards.add(pvscrollPane, folioName);
      displayPanel(folioName, pvTables.size() - 1); // this has to be redone.
    }
  
  public void itemStateChanged(ItemEvent e) {
    JMenuItem source = (JMenuItem)(e.getSource());
    String s = "Item event detected."
      + newline
      + "    Event source: " + source.getText()
      + " (an instance of " + getClassName(source) + ")"
      + newline
                   + "    New state: " 
      + ((e.getStateChange() == ItemEvent.SELECTED) ?
	 "selected":"unselected");
       }

    // Returns just the class name -- no package info.
    protected String getClassName(Object o) {
        String classString = o.getClass().getName();
        int dotIndex = classString.lastIndexOf(".");
        return classString.substring(dotIndex+1);
    }

    public static void main(String[] args) {
        FolioTracker window = new FolioTracker();

        window.setTitle("FolioTracker");
        window.setSize(450, 260);
        window.setVisible(true);
    }
}
