// LDvals.c
//
// this is currently configured to use upper and lower bounds (confidence
// limits on d prime). it used to read one value for LD, so it should be
// fairly straightforward to return to that state. i've noted (hopefully)
// most of the places that should be uncommented.
//
#include <stdio.h>
#include "LDvals.h"
#include "defs.h"

double threshold = THRESHOLD;
int useconfidence = 1;
double upperlimit = .98;
double lowerlimit = .7;


// TO CHANGE TO ONE LD VALUE
// use this array instaed of UPvalues and LOvalues
//
// double LDvalues[MAXSNPS][MAXSNPSPERSNP];
//
double UPvalues[MAXSNPS][MAXSNPSPERSNP];
double LOvalues[MAXSNPS][MAXSNPSPERSNP];

int SNPpositions[MAXSNPS];

int belowTH(int, int, int);
double getUpperlimit(int snp1, int snp2);
double getLowerlimit(int snp1, int snp2);
double getLDvalue(int snp1, int snp2);

//*************************************************************************
// 
// SETTING PARAMETERS
// to change parameters during runtime, call these functions.
//
//*************************************************************************
// setThreshold
//
// sets threshold from outside of this file.
//
int setThreshold (double th) {
  threshold = th;
}


//*************************************************************************
// setUseconfidence
//
// do we want to use confidence bounds or use straight LD?
// not sure if this actually affects what it should affect.
//
// set = 0 or 1
//
int setUseconfidence(int set) {
  useconfidence = set;
}


//*************************************************************************
// setupperlimit
//
int setupperlimit (double up) {
  upperlimit = up;
}

//*************************************************************************
//
// setlowerlimit
//
int setlowerlimit (double lo) {
  lowerlimit = lo;
}
//*************************************************************************
// end setting functions
//

//*************************************************************************
// getSNPposition
//
// returns the position of a given SNP index
//
int getSNPposition(int index) {
  return SNPpositions[index];
}

//*************************************************************************
// BLOCKFINDING FUNCTIONS
//
//

//*************************************************************************
// percentofinfoinLD
//
// returns the % of informative pairs that are in strongLD.
//
double percentofinfoinLD(int snp1, int snp2) {
  int upper, lower;
  int countLD = 0, countinfo = 0;
  int i,j;

  if (snp1 > snp2) {upper = snp1; lower = snp2;}
  else {upper = snp2; lower = snp1
;}

  for (i = lower; i < upper; i++) {
    for (j = i+1; j <= upper; j++) {
      if (getUpperlimit(i,j) < .90)
	countinfo++;
      else if (paircheck(i,j)) {
	countLD++;
	countinfo++;
      }
    }
  }

  return (double)countLD / (double)countinfo;

}

//*************************************************************************
// abovegap
// 
// choose the closest snp which is at least GAP above startSNP.
//
int abovegap(int startSNP) {
  int stoppos;
  int i = startSNP;
  stoppos = getSNPposition(startSNP) + GAP;
  while ((i < SNPcount) && (getSNPposition(i) < stoppos))
    i++;
  return i;
}

//*************************************************************************
// belowgap
//
// choose the closest snp which is at least GAP below startSNP.
//
int belowgap(int startSNP) {
  int stoppos;
  int i = startSNP;
  stoppos = getSNPposition(startSNP) - GAP;
  while ((i > 0) && (getSNPposition(i) > stoppos))
    i--;
  return i;
}

//*************************************************************************
// chooseSNP
//
// choose the farthest snp from start, within maxlength and end.
//
int chooseSNP(int start, int end, int maxlength) {
  int i, stoppos;
  stoppos = getSNPposition(start) + maxlength;
  for (i = start; i <= end; i++) {
    if (getSNPposition(i) > stoppos)
      return i;
  }
  return end;
}

//*************************************************************************
// test
// 
// (i know, a horrible function name.)
//
// returns the percentage that meets the given threshold or other criteria.
//
double test(int snp1, int snp2) {
  int notTH = 0;
  int numsnps, numpairs;
  int upper, lower;
  int i;
  double dnotTH, dnumpairs, percentage;

  if (useconfidence) {
    return percentofinfoinLD(snp1, snp2);
  }
  else {
  if (snp1 > snp2) {upper = snp1; lower = snp2;}
  else {upper = snp2; lower = snp1;}
 
  numsnps = upper - lower + 1;
  numpairs = numsnps * (numsnps - 1) / 2;

  for (i = lower + 1; i <= upper; i++) {
    notTH += belowTH(i, lower, i-1);
  }

  dnotTH = notTH;
  dnumpairs = numpairs;

  percentage = dnotTH / dnumpairs;

  return percentage;
  }
}

//*************************************************************************
// belowTH
//
// this is used for STRAIGHT LD.
//
// returns the number of pairs of SNPs which have LD
// values below the accepted threshold.
// the pairs of SNPs considered are between snp1 and each
// snp from startsnp to endsnp.
// startsnp < endsnp (no check for this)
//
int belowTH(int snp1, int startsnp, int endsnp) {
  int cursnp;
  int count = 0;

  for (cursnp = startsnp; cursnp <= endsnp; cursnp++) {
    if (paircheck(snp1, cursnp))
      count++;
  }

  return count;

}


//*************************************************************************
// paircheck
// 
// is this pair in LD?
// either confidence or straight LD.
//
int paircheck(int snp1, int snp2) {
  if (useconfidence) {
    return (getUpperlimit(snp1, snp2) > upperlimit && getLowerlimit(snp1, snp2) > lowerlimit);
  }
  else
    return (getLDvalue(snp1, snp2) > threshold);
}

//*************************************************************************
//
// TWO AND THREE MARKER BLOCKS
// change the parameters here for different populations.
//

//*************************************************************************
// twocheck
//
// uses parameters for checking a block based on a pair of SNPs.
//
int twocheck(int snp1, int snp2) {
  if (useconfidence) {
    return (getUpperlimit(snp1, snp2) > upperlimit && getLowerlimit(snp1, snp2) > .80);
  }
  else
    return (getLDvalue(snp1, snp2) > threshold);
}

//*************************************************************************
// threecheck
//
// uses parameters for checking a block based on THREE SNPs.
//
int threecheck(int snp1, int snp2) {
  if (useconfidence) {
    return (getUpperlimit(snp1, snp2) > upperlimit && getLowerlimit(snp1, snp2) > .75);
  }
  else
    return (getLDvalue(snp1, snp2) > threshold);
}

//*************************************************************************
// getUpperlimit
//
// returns the upper/lower 5%/95% confidence bounds on Dprime that were
// read from the file.
//
double getUpperlimit(int snp1, int snp2) {
  if (snp1 < snp2) {
    return UPvalues[snp1][snp2 - snp1];
  }
  else if (snp1 == snp2) return 1.00;
  else return UPvalues[snp2][snp1 - snp2];
}

//*************************************************************************
// getLowerlimit
//
double getLowerlimit(int snp1, int snp2) {
  if (snp1 < snp2) {
    return LOvalues[snp1][snp2 - snp1];
  }
  else if (snp1 == snp2) return 1.00;
  else return LOvalues[snp2][snp1 - snp2];
}

//*************************************************************************
// getLDvalue
//
// i think you need to uncomment this if you want to use the LDvalues
// array.
//
// get the LD of snp1 and snp2.
// snp1 and snp2 are the indexes (in SNPpositions) of relevant SNPs.
// snp1 != snp2 but we include a check anyways.

double getLDvalue(int snp1, int snp2) {
  /*
  if (snp1 < snp2) {
    return LDvalues[snp1][snp2];
  }
  else if (snp1 == snp2) {
    return 1.00;
  }
  else return LDvalues[snp2][snp1];
  */
}

//*************************************************************************
// printSNPS
//
void printSNPS() {

  int i;

  for (i =0; i <= SNPcount; i++)
    printf("%d\n",SNPpositions[i]);

}



//*************************************************************************
// UNCOMMENT FOR LD.
//

// prints out a table of LD values

void printLDvalues() {
  /*
  int i, j;

  for (i =0; i <= SNPcount; i++) {
    for (j = 0; j < SNPcount - i; j++) {
      printf ("%.2f\t", LDvalues[i][j]);
      
    }
    printf ("\n");
    
    }*/

}

// reads in LD values from filename
// filename format:
// snp1    snp2    LD

int readLDvalues(char filename[]) {
  /*
  int snp1, snp2;
  double ldval;
  FILE *ldfp;
  int i = 0;
  int lastsnp = 0;
  int snpcount = -1; // in order to get the array to start at 0

  ldfp = fopen(filename, "r");

  while (fscanf(ldfp, "%d %d %lf", &snp1, &snp2, &ldval) != EOF) {
    if (lastsnp != snp1) {
      snpcount++;
      SNPpositions[snpcount]=snp1;
      lastsnp = snp1;
      i = snpcount + 1; // the next item is one step up.
      
    }
    
    LDvalues[snpcount][i] = ldval;
    i++;

    //    printf("%d || %d || %d !!\n", i, snp1, snp2);
  }

  // after this, remember that the last snp is not accounted for.
  snpcount++;
  SNPpositions[snpcount]=snp2;

  fclose(ldfp);
  SNPcount = snpcount;

  printf ("%d\n", snpcount);

  return snpcount;
  */
}

//*************************************************************************
// IF USING CONFIDENCE LIMITS..
//
// readConfvalues
//
// reads in LD values from filename
// filename format:
// snp1    snp2    UP    LO
// should include frequency filter here.
int readConfvalues(char filename[]) {

  int snp1, snp2;
  double upper, lower;
  FILE *ldfp;
  int i = 0;
  int lastsnp = 0;
  int snpcount = -1; // in order to get the array to start at 0

  ldfp = fopen(filename, "r");

  while (fscanf(ldfp, "%d %d %lf %lf", &snp1, &snp2, &lower, &upper) != EOF) {
      if (lastsnp != snp1) {
	snpcount++;
	SNPpositions[snpcount]=snp1;
	lastsnp = snp1;
	i = snpcount + 1; // the next item is one step up.      
      }
    if (snp2 - snp1 < MAXBLOCKLENGTH) {
      

      
      UPvalues[snpcount][i - snpcount] = upper;
      LOvalues[snpcount][i - snpcount] = lower;
      
      i++;
    }
    
    //    printf("%d || %d || %d !!\n", i, snp1, snp2);
  }

  // after this, remember that the last snp is not accounted for.
  snpcount++;
  SNPpositions[snpcount]=snp2;

  fclose(ldfp);
  SNPcount = snpcount;

  printf ("%d\n", snpcount);
  //  printSNPS();

  return snpcount;

}
