/************************************************************************
 * Calculate LD using a variety of metrics				*
 * edited 7/12 to add confidence bounds on D'                           *
 * reads in from a file which also contains the minor allele count      *
 * i took out all the unnecessary calculations to make it as efficient  *
 * as possible.                                                         *
 * compile with: "gcc -O3 -lm -o LD-fast LD-fast.c"                     *
 * this ignores spacings > 200kb                                        *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/stat.h>

#define MAXDIST 200000   // MAXBLOCKLENGTH - do not calc LD for points farther away.

#define MxC 2000
#define MxL MxC + 100
#define True 1
#define False 0

#ifndef Me_Max
#define Me_Max(a,b) ((a) > (b) ? (a) : (b))
#define Me_Min(a,b) ((a) < (b) ? (a) : (b))
#endif

typedef int Boolean ;

struct alleles {
	char a[MxC] ;
} ;

double Chi2=0 ;

int Chromosomes(char *) ;
int Array[3][3] ;

void Clean(char *, int, char *, int) ;
void Three_by_Three(char *, char *, int) ;
void ProgramDescription() ;
void Haplo_Knw(int *) ;
void Haplo_Act(int *, char *, char *, int) ;
void Haplo_Em(int *) ;
void Linkage_Disequilibrium(int *) ;
void Calculate_LD(int, int, int, double *Pos, double *Freq, struct alleles *,
	Boolean, Boolean, Boolean) ;

double calcPDgivenAB(double, double, double, double, double, double, double);
double power(double, int);


int MxS=2000 ;
Boolean Position=False,ChiSquare=False,Bounds=False ;

FILE *fpo=NULL ;

int main (int argc, char *argv[])
{
 int i,k,nm,Num=2,Nc,Nr ;
 char In[512],Out[512],Line[MxL],Rem[512],junk[5][MxC] ;
 char r[100] ;
 double *Pos,*Freq ;
 struct stat stbuf ;
 struct alleles *Snp ;
 Boolean Act=False,Est=True,Head=False,Knw=False,Fin=False; 
 Boolean Fout=False,Remove=False,Separate=False ; 
 FILE *fp=NULL ;


 for ( i=1 ; i < argc ; i++ ) {
   if (argv[i][0] == '-') {
     switch(argv[i][1]) {

       case 'A':                   /* Actual LD */
            Act = True ;
            Est = Knw = False ;
            break ;

       case 'B':                   /* Calculate Confidence Bounds on D' values */
            Bounds = True ;
            break ;

       case 'C':                   /* Calculate Chi-Squared values */
            ChiSquare = True ;
            break ;

       case 'E':                   /* Estimated LD */
            Est = True ;
            Act = Knw = False ;
            break ;

       case 'H':                   /* Print Header */
            Head = True ;
            break ;

       case 'I':                   /* Input File */
            strcpy(In,&argv[i][2]) ;
            Fin = True ;
            break ;

       case 'K':                   /* Known LD (remove double hets)*/
            Knw = True ;
            Act = Est = False ;
            break ;

       case 'O':                   /* Output File */
            strcpy(Out,&argv[i][2]) ;
            Fout = True ;
            break ;

       case 'P':                   /* Write positions rather than distance */
            Position = True ;
            break ;

       case 'R':                   /* Characters to Remove */
            strcpy(Rem,&argv[i][2]) ;
            Remove = True ;
            break ;

       case 'S':                   /* Multiple regions separated by an > */
            Separate = True ;
            break ;

     }
   }
 }

 if (Fin) {
  fp = fopen(In,"r") ;
  if (stat(In,&stbuf) == -1) {
   fprintf(stderr,"\n\nCannot Access Input File %s\n",In) ;
   fprintf(stderr,"Exiting Now......\n\n") ;
   exit(1) ;
  }
 }
 else {
  fprintf(stderr,"\n\nNeed an Input File\n") ;
  fprintf(stderr,"Exiting Now......\n\n") ;
  exit(1) ;
 }

 if (Fout) {
  fpo = fopen(Out,"w") ;
  if (stat(Out,&stbuf) == -1) {
   fprintf(stderr,"\n\nCannot Access Output File %s\n",In) ;
   fprintf(stderr,"Exiting Now......\n\n") ;
   exit(1) ;
  }
 }
 else
  fpo = stdout ;


 if (Head) {
  if (Position)
   fprintf(fpo,"\t\t\t\t  r\t  d\t  D\'\tdelta\t  Q\t rho") ;
  else
   fprintf(fpo,"\t\t\t  r\t  d\t  D\'\tdelta\t  Q\t rho") ;
  if (Bounds)
    fprintf(fpo, "\tLower\tUpper");
  fprintf(fpo, "\n");
  if (!Fin) exit(0) ;
 }

 if (!Act && !Est && !Knw) Est = 1 ;


/************************************************
 * Main Body of the Program			*
 ************************************************/

 Pos  = (double *)malloc(MxS * sizeof(double)) ; 
 Freq = (double *)malloc(MxS * sizeof(double)) ; 
 Snp  = (struct alleles *)malloc(MxS * sizeof(struct alleles)) ;

 Nr = Chromosomes(&Rem[0]) ;
 for (k=0;k<Nr;k++) r[k] = Rem[k] ;

 fp = fopen(In,"r") ;
 Num = Nc = nm = 0 ;

 while (fgets(Line,MxL,fp)) {
  if (4 == sscanf(Line,"%s %s %s %s",junk[0],junk[1],junk[2],junk[3])) {
   Pos[Num]  = atof(junk[0]) ; 
   Freq[Num] = (atof(junk[2])/atof(junk[1])) ;
   strcpy(Snp[Num].a,junk[3]) ;
   if (Num == 0) {
    Nc = Chromosomes(&Snp[Num].a[0]) ;
   }
   else {
    nm = Chromosomes(&Snp[Num].a[0]) ;
    if (Nc != nm) {
     fprintf(stderr,"\n\nError in File %s at Line %d\n",In,Num+1) ;
     fprintf(stderr,"Wrong number of chromosomes\n") ;
     fprintf(stderr,"Exiting Now....\n\n\n") ;
     exit(1) ;
    }
   }
   Num ++ ;
   if (MxS <= Num) {
    MxS += 1000 ;
    Pos  = (double *)realloc(Pos, MxS * sizeof(double)) ; 
    Freq = (double *)realloc(Freq,MxS * sizeof(double)) ; 
    Snp  = (struct alleles *)realloc(Snp,MxS * sizeof(struct alleles)) ; 
   }
  }
  else {

   for (i=0;i<Num;i++) Clean(&Snp[i].a[0],Nc,&r[0],Nr) ;
   Calculate_LD(Nc,Nr,Num,Pos,Freq,Snp,Act,Est,Knw) ;
   Num = Nc = nm = 0 ;

   if (Separate) fprintf(fpo,">\n") ;
  }
 }


 for (i=0;i<Num;i++) Clean(&Snp[i].a[0],Nc,&r[0],Nr) ;
 Calculate_LD(Nc,Nr,Num,Pos,Freq,Snp,Act,Est,Knw) ;
 if (Separate) fprintf(fpo,">\n") ;

 exit(0) ;

}




/******************************************************************
 *  SS   U  U  BBB   RRR    OO   U  U TTTTT III  N N   EEEE   SS  *
 * S  S  U  U  B  B  R  R  O  O  U  U   T    I   NN N  E     S  S *
 * S     U  U  B  B  R  R  O  O  U  U   T    I   N  N  E     S    *
 *  S    U  U  BBB   RRR   O  O  U  U   T    I   N  N  EEE    S   *
 *   S   U  U  B  B  R R   O  O  U  U   T    I   N  N  E       S  *
 *    S  U  U  B  B  R  R  O  O  U  U   T    I   N  N  E        S *
 * S  S  U  U  B  B  R  R  O  O  U  U   T    I   N  N  E     S  S *
 *  SS    UU   BBB   R  R   OO    UU    T   III  N  N  EEEE   SS  *
 ******************************************************************/


/********************************************************
 * Calculate LD for all SNP pairs.			*
 ********************************************************/
void Calculate_LD(int Nc,int Nr,int Num,double *Pos, double *Freq,
	struct alleles *Snp, Boolean Act, Boolean Est, Boolean Knw)
{
 int i,j,k,Fa[4] ;
 char a[MxC],b[MxC] ;

 for (i=0;i<Num-1;i++) {
  
   if (Pos[i+1] - Pos[i] > MAXDIST)
     fprintf(fpo,"%d\t%d 0 0 0 0 0 0 0 0 0 1\n",(int)Pos[i],(int)Pos[i+1]) ;
   else {
     for (k=0;k<Nc;k++) a[k] = Snp[i].a[k] ;
     
     //for (j=i+1; j<Num ; j++) {
     for (j=i+1; (j<Num && (Pos[j] -Pos[i] < MAXDIST)); j++) {
       if (Position)
	 fprintf(fpo,"%d\t%d\t",(int)Pos[i],(int)Pos[j]) ;
       else
	 fprintf(fpo,"%d\t",(int)(Pos[j]-Pos[i])) ;
       //       fprintf(fpo,"%.2f\t%.2f\t",Freq[i],Freq[j]) ;
       
       //if (Pos[j] -Pos[i] < MAXDIST) {
       
       for (k=0;k<Nc;k++) b[k] = Snp[j].a[k] ;
       
       if (Freq[i] == 0 || Freq[j] == 0)
	 fprintf(fpo,"\n") ;
       else 
	 if (!Act) Three_by_Three(&a[0],&b[0],Nc) ;
       
       if (Est) Haplo_Em(&Fa[0]) ;
       else if (Knw) Haplo_Knw(&Fa[0]) ;
       else if (Act) Haplo_Act(&Fa[0],&a[0],&b[0],Nc) ;
       
       Linkage_Disequilibrium(&Fa[0]) ;
     }
   // else fprintf(fpo,"0 0 0 0 0 0 0 1\n") ;

   }
 }
 
}

/********************************************************
 * Count the number of chromosomes.			*
 ********************************************************/
int Chromosomes(char *chrom)
{
 int i=0 ;
 while (chrom[i] != '\0') i++ ;
 return(i) ;
}

/********************************************************
 * Remove unwanted terms in allele info and change	*
 * markers to 0's and 1's.				*
 ********************************************************/
void Clean(char *Snp, int Nc, char *r, int Nr)
{
 int i,j,na=0,ns=2 ;
 char All[10],Sub[2] ;
 Boolean Add ;

 Sub[0] = '0' ;
 Sub[1] = '1' ;

 for (i=0;i<Nr;i++) {
  for (j=0;j<Nc;j++)
   if (Snp[j] == r[i]) Snp[j] = '-' ;
 }

 for (i=0;i<Nc;i++) {
  if (Snp[i] != '-') {
   Add = True ;
   for (j=0;j<na;j++)
    if (All[j] == Snp[i]) Add = False ;
   if (Add) {
    All[na] = Snp[i] ;
    na ++ ;
   }
  }
 }

 if (2 < na) {
  fprintf(stderr,"\n\nMore Than Two Alleles %s\n",Snp) ;
  fprintf(stderr,"Exiting Now....\n\n\n") ;
  exit(1) ;
 }

 for (i=0;i<ns;i++) {
  for (j=0;j<na;j++) {
   if (All[j] == Sub[i]) {
    Sub[i] = Sub[ns-1] ;   
    All[j] = All[na-1] ;   
    na -- ;
    ns -- ;
    i -- ;
    break ;
   }
  }
 }
 
 for (j=0;j<na;j++) {
  for (i=0;i<Nc;i++)
   if (Snp[i] == All[j]) Snp[i] = Sub[j] ;
 }

}


/********************************************************
 * Create the 3x3 table with allele observations like:	*
 *							*
 * 		AA	Aa	aa			*
 *	     ______________________			*
 *	BB  |	n1	n2	n3 |			*
 *	Bb  |	n4	n5	n6 |			*
 *	bb  |___n7______n8______n9_|			*
 *							*
 ********************************************************/
void Three_by_Three(char *a, char *b, int Nc)
{
 int i,j,k ;

 for (i=0;i<3;i++)
  for (j=0;j<3;j++) Array[i][j] = 0 ;

 for (k=0;k<Nc;k+=2) {
  i = j = -1 ;
  if (a[k] == '0' && a[k+1] == '0') j = 0 ; 
  if (a[k] == '0' && a[k+1] == '1') j = 1 ;
  if (a[k] == '1' && a[k+1] == '0') j = 1 ;
  if (a[k] == '1' && a[k+1] == '1') j = 2 ;

  if (b[k] == '0' && b[k+1] == '0') i = 0 ; 
  if (b[k] == '0' && b[k+1] == '1') i = 1 ; 
  if (b[k] == '1' && b[k+1] == '0') i = 1 ; 
  if (b[k] == '1' && b[k+1] == '1') i = 2 ; 

  if (i != -1 && j != -1) Array[i][j] ++ ;
 }

}


/********************************************************
 * Get the known haplotypes.				*
 *							*
 * 	x = AB = 00 freq				*
 * 	y = Ab = 01 freq				*
 * 	z = aB = 10 freq				*
 *	w = ab = 11 freq				*
 *							*
 ********************************************************/
void Haplo_Knw(int *Fa)
{
 int x,y,z,w ;

 x = 2*Array[0][0] + Array[0][1] + Array[1][0] ;
 y = 2*Array[2][0] + Array[1][0] + Array[2][1] ;
 z = 2*Array[0][2] + Array[0][1] + Array[1][2] ;
 w = 2*Array[2][2] + Array[1][2] + Array[2][1] ;

 Fa[0] = x ;
 Fa[1] = y ;
 Fa[2] = z ;
 Fa[3] = w ;

}


/********************************************************
 * Get the actual haplotypes.				*
 *							*
 * 	x = AB = 00 freq				*
 * 	y = Ab = 01 freq				*
 * 	z = aB = 10 freq				*
 *	w = ab = 11 freq				*
 *							*
 ********************************************************/
void Haplo_Act(int *Fa, char *a, char *b, int Nc)
{
 int i,x,y,z,w ;

 x = y = z = w = 0 ;

 for (i=0;i<Nc;i++) {
  if (a[i] == '0' && b[i] == '0') x ++ ;
  if (a[i] == '0' && b[i] == '1') y ++ ;
  if (a[i] == '1' && b[i] == '0') z ++ ;
  if (a[i] == '1' && b[i] == '1') w ++ ;
 }

 Fa[0] = x ;
 Fa[1] = y ;
 Fa[2] = z ;
 Fa[3] = w ;

}

double findlnandmult(double pnumber, double count) {
  if (count == 0)
    return 0;
  else return count * log(pnumber);
}

int likelihoodshouldbezero(double pnumber, double count) {
  if (pnumber == 0)
    if (count > 0)
      return 1;
  return 0;
}

double calclike(double Dprime, double pA, double pB, double AB, double Ab, double aB, double ab, double DH, double s, double t, double adjust) {
  double pAB, pAb, paB, pab;
  double lnlike, like;

  pAB = Dprime * s + t;
  pAb = pA - pAB;
  paB = pB - pAB;
  pab = 1 - pAB - pAb - paB;

  if (likelihoodshouldbezero(pAB, AB) || likelihoodshouldbezero(pAb, Ab) ||
      likelihoodshouldbezero(paB, aB) || likelihoodshouldbezero(pab, ab) ||
      likelihoodshouldbezero(pAB*pab + pAb*paB, DH))
    return 0;
  else { 
      
    lnlike = findlnandmult(pAB, AB)+ findlnandmult(pAb, Ab)+ 
      findlnandmult(paB, aB) + findlnandmult(pab, ab) + 
      findlnandmult(pAB*pab + pAb*paB, DH);
    like = exp(lnlike - adjust);
    return like;
  }
  
}


double calclnlike(double Dprime, double pA, double pB, double AB, double Ab, double aB, double ab, double DH, double s, double t) {
  double pAB, pAb, paB, pab;
  double lnlike;

  pAB = Dprime * s + t;
  pAb = pA - pAB;
  paB = pB - pAB;
  pab = 1 - pAB - pAb - paB;

  lnlike = findlnandmult(pAB, AB)+ findlnandmult(pAb, Ab)+ 
    findlnandmult(paB, aB) + findlnandmult(pab, ab) + 
    findlnandmult(pAB*pab + pAb*paB, DH);
  return lnlike;
}

int calcConfidenceLimits(double returnval[], double emAB, double emDp,
			     double NA, double NB, double Na, double Nb) {
 
  double AB, Ab, aB, ab, unknown, total;
  double loglikelihood, likelihood;
  double Dprime, step = .01, s, t;
  double normalize;
  double pA, pB, pa, pb, pAB, pAb, paB, pab;
  double integral, oldval, newval;
  double swap;
  double maxval, someval, preexp;

  AB = (double) 2*Array[0][0] + Array[0][1] + Array[1][0] ;
  Ab = (double) 2*Array[2][0] + Array[1][0] + Array[2][1] ;
  aB = (double) 2*Array[0][2] + Array[0][1] + Array[1][2] ;
  ab = (double) 2*Array[2][2] + Array[1][2] + Array[2][1] ;
  unknown = (double) Array[1][1];

  total = AB + Ab + aB + ab + 2 *unknown;

  NA = (double) 2*(Array[0][0] + Array[1][0] + Array[2][0]) +
    Array[0][1] + Array[1][1] + Array[2][1];

  NB = (double) 2*(Array[0][0] + Array[0][1] + Array[0][2]) +
    Array[1][0] + Array[1][1] + Array[1][2];

  Na = total - NA;
  Nb = total - NB;

  if (NA * NB / total > emAB) {
    // swap A and a
    swap = NA;
    NA = Na;
    Na = swap;
	     
    swap = AB;
    AB = aB; 
    aB = swap;
    
    swap = Ab;
    Ab = ab; 
    ab = swap;
  }

  pA = NA / total;
  pB = NB / total;
  pa = Na / total;
  pb = Nb / total;

  s = Me_Min(pA*pb, pa*pB);
  t = pA*pB;

  maxval = calclnlike(0, pA, pB, AB, Ab, aB, ab, unknown, s, t);

  // find the maxval first
  for (Dprime = 0; Dprime < 1; Dprime += step) {
    pAB = Dprime * s + t;
    pAb = pA - pAB;
    paB = pB - pAB;
    pab = pa - paB;
    if (! (likelihoodshouldbezero(pAB, AB) || likelihoodshouldbezero(pAb, Ab) ||
	likelihoodshouldbezero(paB, aB) || likelihoodshouldbezero(pab, ab) ||
	   likelihoodshouldbezero(pAB*pab + pAb*paB, unknown))) {
      preexp = calclnlike(Dprime, pA, pB, AB, Ab, aB, ab, unknown, s, t);
      if (preexp > maxval)
	maxval = preexp;
    }
  }

  normalize = 0;

  for (Dprime = 0; Dprime < 1; Dprime += step) {
    pAB = Dprime * s + t;
    pAb = pA - pAB;
    paB = pB - pAB;
    pab = pa - paB;
    someval = calclike(Dprime, pA, pB, AB, Ab, aB, ab, unknown, s, t, maxval);
    normalize += someval;
  }

  // find lower bound
  integral = 0;
  Dprime = 0;
  pAB = Dprime * s + t;
  pAb = pA - pAB;
  paB = pB - pAB;
  pab = pa - paB;
  integral = calclike(Dprime, pA, pB, AB, Ab, aB, ab, unknown, s, t, maxval) / normalize;

  while (integral < .05) {
    Dprime += step;
    pAB = Dprime * s + t;
    pAb = pA - pAB;
    paB = pB - pAB;
    pab = pa - paB;
    oldval = calclike(Dprime, pA, pB, AB, Ab, aB, ab, unknown, s, t,  maxval) / normalize;
    integral += oldval;
  }
  returnval[0] = Dprime;


  integral = 0;
  Dprime = 1;
  pAB = Dprime * s + t;
  pAb = pA - pAB;
  paB = pB - pAB;
  pab = pa - paB;
  integral = calclike(Dprime, pA, pB, AB, Ab, aB, ab, unknown, s, t, maxval) / normalize;

  while (integral < .05) {
    Dprime -= step;
    pAB = Dprime * s + t;
    pAb = pA - pAB;
    paB = pB - pAB;
    pab = pa - paB;
    oldval = calclike(Dprime, pA, pB, AB, Ab, aB, ab, unknown, s, t,  maxval) / normalize;

    integral += oldval;
  }

  returnval[1] = Dprime;



}

double power(double x, int y) {
  int i;
  double answer = 1;
  
  for (i = 0; i < y; i++) {
    answer = answer * x;
  }
  
  return answer;
}
  
  




/********************************************************
 * Calculate various LD metrics using the following	*
 * definition.						*
 *							*
 *	  B	  b					*
 * A	P[AB]	P[Ab]	P[A]				*
 * a	P[aB]	P[ab]	P[a]				*
 * 	P[B]	P[b]	1				*
 *							*
 * where SNP 1 has alleles A/a and SNP 2 has alleles	*
 * B/b, P[AB] means the number of haplotype AB obser-	*
 * vations and P[A]=P[AB]+P[Ab].			*
 ********************************************************/
void Linkage_Disequilibrium(int *Fa)
{
 double PAB,PAb,PaB,Pab,PA,Pa,PB,Pb,total ;
 double D,r,d,Dprim,delta,Q,rho,Dpconfidence[2] ;
 // added D prime confidence bounds at 5% and 95% - catfoo
 double Large = 1000 ;

 if (Fa[1] < Fa[2]) {
  PAB = (double)Fa[0] ;
  PAb = (double)Fa[1] ;
  PaB = (double)Fa[2] ;
  Pab = (double)Fa[3] ;
  PA  = (double)(Fa[0] + Fa[1]) ;
  Pa  = (double)(Fa[2] + Fa[3]) ;
  PB  = (double)(Fa[0] + Fa[2]) ;
  Pb  = (double)(Fa[1] + Fa[3]) ;
 }
 else {
  PAB = (double)Fa[0] ;
  PAb = (double)Fa[2] ;
  PaB = (double)Fa[1] ;
  Pab = (double)Fa[3] ;
  PA  = (double)(Fa[0] + Fa[2]) ;
  Pa  = (double)(Fa[1] + Fa[3]) ;
  PB  = (double)(Fa[0] + Fa[1]) ;
  Pb  = (double)(Fa[2] + Fa[3]) ;
 }

 if (PA == 0 || Pa == 0 || PB == 0 || Pb == 0) {
  fprintf(fpo,"0 0 0 0 0 0 0 1\n") ;
  return ;
 }

 //   fprintf(fpo,"%d\t%d\t%d\t%d\t", (int)PAB, (int)PAb, (int)PaB, (int)Pab);
 
/************************************************************************
 * From "Disequilibrium measures for fine mapping" by Devlin and Risch	*
 * [1995] that defines D on page 313. Also given on page 113 of		*
 * "Genetic Data Analysis" by Bruce Weir.				*
 ************************************************************************/
 // D	= (PAB*Pab-PAb*PaB) ;

/************************************************************************
 * From "Genetic Data Analysis II" by Bruce Weir that defines r on 	*
 * page 137 (also equal to capital delta and given in "Disequilibrium	*
 * measures for fine mapping" by Devlin and Risch [1995]).		*
 ************************************************************************/
 // r	= (D)/sqrt(PA*Pa*PB*Pb) ;

/************************************************************************
 * From "Disequilibrium measures for fine mapping" by Devlin and Risch	*
 * that defines d in Table 3						*
 ************************************************************************/
 // d	= (D)/(PB*Pb) ;

/************************************************************************
 * From "Disequilibrium measures for fine mapping" by Devlin and Risch	*
 * [1995] that defines D' in Table 3 (I think that this equation is not	*
 * right though). Also given on page 120 of "Genetic Data Analysis" by	*
 * Bruce Weir (I think that this is the right equation).		*
 ************************************************************************/
 // if (0 < D)  Dprim = (D)/Me_Min(PA*Pb,Pa*PB) ;
 // else        Dprim = (-D)/Me_Min(PA*PB,Pa*Pb) ; 

/************************************************************************
 * From "Disequilibrium measures for fine mapping" by Devlin and Risch	*
 * [1995] that defines lower case delta in Table 3			*
 ************************************************************************/
 // if (Pab != 0) delta = (D)/(PB*Pab) ;
 // else delta = Large ;

/************************************************************************
 * From "Disequilibrium measures for fine mapping" by Devlin and Risch	*
 * that defines Q in Table 3						*
 ************************************************************************/
 // Q	= (D)/(PAB*Pab+PAb*PaB) ;


/************************************************************************
 * From "Evaluation of linkage disequilibrium measures" by Ennis et al.	*
 * that defines the association rho in Table 1				*
 ************************************************************************/
 // rho	= (D)/(PA*Pb) ;
 // if (1 < rho || rho < -1)
 //  rho	= (D)/(PB*Pa) ;


/************************************************************************
 * From "The Structure of Haplotype Blocks in the Human Genome" by	*
 * Gabriel et al. [2002] that uses confidence bounds on D prime.        *
 * Description in Materials and Methods supplement online, details      *
 * obtained via personal email to Mark Daly.                            *
 ************************************************************************/
 // total = PAB + PAb + PaB + Pab;
 calcConfidenceLimits(Dpconfidence, PAB, Dprim, PA, PB, Pa, Pb);




// if (rho > 1 || rho < -1) fprintf(stderr,"%f\t%f\t%f\t%f\n",PAB,PAb,PaB,Pab) ;
// printf("%5.2f\t",D) ;
 /* fprintf(fpo,"%5.2f\t",r) ;
 fprintf(fpo,"%5.2f\t",d) ;
 fprintf(fpo,"%5.2f\t",Dprim) ;
 if (delta < -999)
  fprintf(fpo," -NaN\t") ;
 else if (999 < delta)
  fprintf(fpo,"  NaN\t") ;
 else
  fprintf(fpo,"%5.2f\t",delta) ;
 fprintf(fpo,"%5.2f\t",Q) ;
 fprintf(fpo,"%5.2f",rho) ;

 if (ChiSquare)
  fprintf(fpo,"\t%5.3f",Chi2) ;
 */
 // if (Bounds)
   fprintf(fpo, "\t%5.3f\t%5.3f", Dpconfidence[0], Dpconfidence[1]);

 fprintf(fpo,"\n") ;

 
}




/********************************************************
 * Use the EH algorithm to calculate the expected 	*
 * number of haplotypes assuming that the input array	*
 * contains the following elements:			*
 *							*
 * [0,0] = # AA-BB observations				*
 * [0,1] = # Aa-BB observations				*
 * [0,2] = # aa-BB observations				*
 * [1,0] = # AA-Bb observations				*
 * [1,1] = # Aa-Bb observations				*
 * [1,2] = # aa-Bb observations				*
 * [2,0] = # AA-bb observations				*
 * [2,1] = # Aa-bb observations				*
 * [2,2] = # aa-bb observations				*
 *							*
 * Observed values:					*
 *	c1 = AB num					*
 *	c2 = Ab num					*
 *	c3 = aB num					*
 *	c4 = ab num					*
 *	c5 = 0.5 * Unknown num				*
 ********************************************************/
void Haplo_Em(int *Fa)
{
 double x,y,z,w,c1,c2,c3,c4,c5,t1,t2,t3,t4,oldlogL,logL ;
 double logL0,f1,f2,x0,y0,z0,w0,total ;

 x = y = z = w = .25 ; /* initial guess for haplo freqs */

 c1 = t1 = 2*Array[0][0]+Array[0][1]+Array[1][0] ;
 c2 = t2 = 2*Array[2][0]+Array[1][0]+Array[2][1] ;
 c3 = t3 = 2*Array[0][2]+Array[0][1]+Array[1][2] ;
 c4 = t4 = 2*Array[2][2]+Array[1][2]+Array[2][1] ;
 c5 = Array[1][1] ;
 
 total = c1 + c2 + c3 + c4 + (2 * c5) ;	/* Number of SNP Pairs */

 oldlogL = -999999999. ;
 logL  = c1*log(x) + c2*log(y) + c3*log(z) ;
 logL += c4*log(w) + c5*log(2*x*w+2*y*z) ;
 f1 = (c1+c2+c5)/(c1+c2+c3+c4+2*c5) ;
 f2 = (c1+c3+c5)/(c1+c2+c3+c4+2*c5) ;

 x0 = f1*f2 ;
 y0 = f1*(1-f2) ;
 z0 = (1-f1)*f2 ;
 w0 = (1-f1)*(1-f2) ;
 if (x0 == 0 || y0 == 0 || z0 == 0 || w0 == 0) {
  Fa[0] = Fa[1] = Fa[2] = Fa[3] = 0 ;
  return ;
 }
 logL0  = c1*log(x0) + c2*log(y0) + c3*log(z0) ;
 logL0 += c4*log(w0) + c5*log(2*x0*w0+2*y0*z0) ;

 while (0.000000001 < abs(logL-oldlogL)) {
  t1 = c1 + c5*x*w/(x*w+y*z) ;
  t2 = c2 + c5*y*z/(x*w+y*z) ;
  t3 = c3 + c5*y*z/(x*w+y*z) ;
  t4 = c4 + c5*x*w/(x*w+y*z) ;
  x = t1/(t1+t2+t3+t4) ;
  y = t2/(t1+t2+t3+t4) ;
  z = t3/(t1+t2+t3+t4) ;
  w = t4/(t1+t2+t3+t4) ;
  oldlogL = logL ;
  logL = 0 ;
  if (0 < x) logL += (c1 * log(x)) ;
  if (0 < y) logL += (c2 * log(y)) ;
  if (0 < z) logL += (c3 * log(z)) ;
  if (0 < w) logL += (c4 * log(w)) ;
  if (0 < x*w+y*z) logL += c5*log(2*x*w+2*y*z) ;
 }

 Fa[0] = (int)((x * total) + 0.5) ;
 Fa[1] = (int)((y * total) + 0.5) ;
 Fa[2] = (int)((z * total) + 0.5) ;
 Fa[3] = (int)((w * total) + 0.5) ;

 Chi2  = 2 * (logL - logL0) ;

 return ;

}


